

"""
    Optimizer

Composite type holding the method and options for an optimizer, concretized on instantiation of a Problem.
"""
struct Optimizer
  method
  options
end

struct OptHistory{T}
  state_vector::Vector{T}
  gradients::Vector{T}
  objectives::Vector{T}
end

"""
  Problem{T}

Composite structure representing a lasso optimization problem.

# Fields:
- `targets::Dict{Int, AbstractOptTarget}`  # Dictionary containing the optimization targets, indexed by an integer value
- `variables::Dict{Symbol, AbstractOptVariable}`  # Dictionary containing the optimization variables, indexed by variable symbol
- `state_vector::Vector{T}`  # Vector holding the value of the objective function and possibly gradient
- `optimizer::Optimizer`  # Optimization method
- `equilibria::Dict{Symbol, AbstractOptEquilibrium}` # A vmecData object representing the base eq, used for output
- `objective_counter::Int` # An int to keep track of how many objective function evaluations have occurred 
- `gradient_counter::Int` # An int to keep track of the number of gradient evaluations
- `status_file::String` #A file name to write status output to
- `gradient_file::String` #A file name to write gradient output to
"""
mutable struct Problem{T}
  targets::Dict{Int, Target}
  state_vector::Vector{OptimizationVariable}
  optimizer::Optimizer
  equilibrium_wrappers::Dict{UInt, EquilibriumWrapper}
  derived_geometries::Dict{UInt, DerivedGeometryWrapper}
  objective_counter::Int
  gradient_counter::Int
  history::Vector{OptHistory{T}}
  objective::T
  status_file::String
  gradient_file::String
  Problem() = Problem{Float64}()
  Problem{T}() where {T} = new{T}(Dict{Int, Target}(),
                                  Vector{OptimizationVariable}(undef, 0),
                                  Optimizer(nothing, nothing),
                                  Dict{UInt, EquilibriumWrapper}(),
                                  Dict{UInt, Expr}(),
                                  0, 0, Vector{OptHistory{T}}(undef, 0),
                                  convert(T, Inf), "status.txt", "gradients.txt")
  Problem{T}(a, b, c, d, e, f, g, h, i, j, k) where {T} = new(a, b, c, d, e, f, g, h, i, j, k)
end
#=
"""
    Problem(data_type::Type=Float64; status_file::String="status.txt", gradient_file::String="gradients.txt")
"""
function Problem(state_vector::Vector{T};
                 status_file::String="status.txt",
                 gradient_file::String="gradients.txt",
                ) where {T}
  targets = Dict{Int, AbstractOptTarget}()
  variables = Dict{Symbol, AbstractOptVariable}()
  optimizer = Optimizer()
  equilibria = Dict{Type, EquilibriumWrapper}()
  objective_counter = 0
  gradient_counter = 0
  history = Vector{OptHistory{T}}(undef, 0) 
  objective = convert(data_type, Inf)
  return Problem{T}(targets, variables, state_vector, 
                    optimizer, equilibria, objective_counter, gradient_counter,
                    history, objective, status_file, gradient_file)
end


function set_state_vector!(prob::Problem{T}) where {T}
  state_vector = Vector{T}(undef, length(prob.variables))
  for (i, var) in enumerate(prob.variables)
    state_vector[i] = var.second.value
  end
  prob.state_vector = copy(state_vector)
end
=#
