function optimize(inputFile::String)
end

function optimize_threads(prob::Problem)
  prob.objective_counter = 0
  prob.gradient_counter = 0
  initial_state_vector = state_vector_values(prob)
  Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob)),
                 initial_state_vector, prob.optimizer.method, prob.optimizer.options)
end

function objective_gradient!(F,
                             G,
                             θ::AbstractVector,
                             prob::Problem;
                            )
  adjust_state_vector!(prob, θ)
  obj, grad = (Inf, nothing)
  if !isnothing(F)
    @debug "State vector before iteration $(prob.objective_counter): $(transpose(θ))"
    obj = compute_objective(prob)
    adjust_objective!(prob, obj)
    write_status_file(prob)
    prob.objective_counter += 1
  end
  if !isnothing(G)
    if !isnan(obj)
      grad = compute_gradient(prob)
      @debug "Grad vector: $(transpose(grad))"
      copyto!(G, grad)
      write_gradients_file(prob, grad)
      update_history!(prob, OptHistory(θ, grad, [prob.objective]))
      prob.gradient_counter += 1
    else
      copyto!(G, fill(NaN, length(x)))
    end
  end
  
  if !isnothing(F)
    return obj
  end
  nothing
end

function compute_objective(prob::Problem;
                          )
    valid_eqs = fill(true, length(prob.equilibrium_wrappers))
    j = 0
    for i in eachindex(prob.equilibrium_wrappers)
      j += 1
      translate!(prob.equilibrium_wrappers[i], prob.state_vector)
      eq = compute_equilibrium(prob.equilibrium_wrappers[i])
      if !isnothing(eq)
        update_equilibrium!(prob.equilibrium_wrappers[i], eq)
      else
        valid_eqs[j] = false
      end
    end

    if all(valid_eqs)
        compute_targets!(prob)
        F = least_squares_objective(prob)
        return F
    else
        return NaN
    end
end

function compute_gradient(prob::Problem{T};
                         ) where {T}
    G = zeros(T, length(prob.state_vector))
    for grad_index in eachindex(G)
        local_prob = deepcopy(prob)
        G[grad_index] = evaluate_fd_gradient(grad_index, local_prob)
    end
    return G
end

#No backtrack_step input option
function evaluate_fd_gradient(grad_index::Int,
                              prob::Problem;
                             )
    diff_type = prob.state_vector[grad_index].diff_type
    if diff_type === :CenteredDiff
        return centered_diff(grad_index, prob)
    elseif diff_type === :ForwardDiff
        return forward_diff(grad_index, prob)
    else
        error("Method $(diff_type) not a valid finite differencing method")
    end
end

function perturbed_equilibrium(grad_index::Int,
                               perturbation::T,
                               prob::Problem;
                               iter_limit::Int=5,
                               backtrack_step=0.5,
                              ) where{T}
    eq = nothing
    initial_value = prob.state_vector[grad_index].value
    #Just a quick something something. Need to generalized for multiple equilibira
    wrapper_key = prob.state_vector[grad_index].equilibrium_key
    wrapper = prob.equilibrium_wrappers[wrapper_key]
    iter = 0
    while isnothing(eq) && iter <= iter_limit
        if iter > 0
            @warn "Performing backtracking to recompute equilibrium, consider adjusting step size"
        end
        perturbation = backtrack_step^iter .* perturbation
        perturbed_value = initial_value + perturbation
        @debug "grad index: $grad_index, value: $perturbed_value"
        adjust_state_vector!(prob, grad_index, perturbed_value)
        #Just a quick something something. Need to generalized for multiple equilibira
        @debug "new value: $(prob.state_vector[grad_index].value)"
        translate!(wrapper, prob.state_vector)
        
        eq = compute_equilibrium(wrapper)
        iter += 1
    end

    if !isnothing(eq)
        update_equilibrium!(wrapper, eq)
        compute_targets!(prob, wrapper_key = wrapper_key)
        return least_squares_objective(prob), perturbation
    else
        return nothing, perturbation
    end
end

function forward_diff(grad_index::Int,
                      prob::Problem;
                      backtrack_step=0.1,
                     )
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    perturbation = opt_var.finite_diff_step
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
    @debug "Computing at most $iter_limit attempts for derivative"
    @debug "Objective: $(prob.objective)"
    target, perturbation = perturbed_equilibrium(grad_index, perturbation, prob, 
                                                 backtrack_step=backtrack_step)
    opt_var.finite_diff_step = abs(perturbation)
    opt_var.value = initial_value
    return !isnothing(target) ? (target - prob.objective) / opt_var.finite_diff_step : zero(typeof(prob.objective))
end

function centered_diff(grad_index::Int,
                       prob::Problem{T};
                       backtrack_step=0.1,
                      ) where {T}
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    compute_backward, compute_forward = (true, true)
    # TODO change the target values to match the types of actual targets
    target_values = Vector{T}(undef, 2)
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    @debug "Finite difference step: $finite_diff_step"
    while compute_backward || compute_forward
        # Do the negative step
        if compute_backward
            perturbation = -finite_diff_step
            iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
            @debug "Computing at most $iter_limit attempts for derivative"
            targets, perturbation = perturbed_equilibrium(grad_index, perturbation, prob,
                                                          backtrack_step=backtrack_step,
                                                          iter_limit=iter_limit)
            opt_var.value = initial_value
            target_values[1] = !isnothing(targets) ? targets : zero(typeof(prob.objective))
            @debug "$(abs(perturbation)) $(finite_diff_step)"
            if abs(perturbation) != finite_diff_step
                if abs(perturbation) >= min_step_size
                    finite_diff_step = copy(abs(perturbation))
                    compute_forward = true
                    @debug "Finite difference step size adjust to $finite_diff_step"
                else
                    finite_diff_step = copy(min_step_size)
                    compute_forward = false
                    target_values = zeros(eltype(prob.state_vector[1].value), 2)
                end
            end
           compute_backward = false
        end

        # Do the positive step
        if compute_forward
            perturbation = finite_diff_step
            iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
            @debug "Computing at most $iter_limit attempts for derivative"
            targets, perturbation =  perturbed_equilibrium(grad_index, perturbation, prob,
                                                           backtrack_step=backtrack_step,
                                                           iter_limit=iter_limit)
            opt_var.value = initial_value
            target_values[2] = !isnothing(targets) ? targets : zero(typeof(prob.objective))
            @debug "$(abs(perturbation)) $(finite_diff_step)"
            if abs(perturbation) != finite_diff_step 
                if abs(perturbation) >= min_step_size
                    finite_diff_step = copy(abs(perturbation))
                    compute_backward = true
                    @debug "Finite difference step size adjust to $finite_diff_step"
                else
                    finite_diff_step = copy(min_step_size)
                    compute_backward = false
                    target_values = zeros(T, 2)
                end
            end
            compute_forward = false
        end
    end

    @debug "finite_diff_step after backtracking: $(opt_var.finite_diff_step)"
    opt_var.value = initial_value
    return 0.5 * (target_values[2] - target_values[1]) / finite_diff_step
end

function least_squares_objective(prob::Problem{T}) where {T}
    obj = zero(T)
    for (index, target) in prob.targets
        #length(target.label) < 25 ? target_label = target.label * repeat(" ", 25-length(target.label)) : target_label = target.label
        #target_string = @sprintf "%.3E" target.target_value
        #value_string = @sprintf "%.5E" target.current_value
        #sigma_string = @sprintf "%.3E" target.target_weight
        #println(target_label,": target ",target_string," value ",value_string," sigma ",sigma_string)
        if isa(target.target_value, Tuple) || isa(target.target_value, AbstractVector)
            for i = 1:length(target.target_value)
                obj += (target.target_value[i] - target.current_value[i])^2 / (target.target_weight[i]^2)
            end
        else
            obj += (target.target_value - target.current_value)^2 / (target.target_weight^2)
        end
    end
    objective_string = @sprintf "%.5E" obj
    #println("Objective value: ",objective_string)
    return obj
end

function write_status_file(prob::Problem)
    if prob.objective_counter == 0
        fid = open(prob.status_file,"w")
        write(fid, "#iteration;target_name;target;value;sigma;(v/s)^2\n")
    else
        fid = open(prob.status_file,"a")
    end
    istring = @sprintf "%3i" prob.objective_counter
    for (index, target) in prob.targets 
        length(target.label) < 25 ? target_label = target.label * repeat(" ", 25-length(target.label)) : target_label = target.label
        target_string = @sprintf "%.7E" target.target_value
        value_string = @sprintf "%.7E" target.current_value
        sigma_string = @sprintf "%.7E" target.target_weight
        vos_string = @sprintf "%.7E" abs(target.current_value - target.target_value)^2/target.target_weight^2
        write(fid,istring*"; "*target_label*"; ")
        write(fid,target_string*"; "*value_string*"; "*sigma_string*"; "*vos_string*"\n")
    end
    close(fid)
end

function write_gradients_file(prob::Problem,
                              gradient::AbstractVector;
                             )
    p_iter = prob.gradient_counter
    if p_iter <= 1
        return
    end
    if p_iter == 2
        fid = open(prob.gradient_file,"w")
        write(fid, "#iteration;varIndex;old_value;new_value;gradient;alpha*P;step\n")
    else
        fid = open(prob.gradient_file,"a")
    end
    nvals = length(prob.state_vector)
    istring = @sprintf "%3i" p_iter
    for j in 1:nvals
        jstring = @sprintf "%3i" j
        nv_string = @sprintf "%.7E" prob.state_vector[j].value
        old_value = prob.history[p_iter-1].state_vector[j]
        ov_string = @sprintf "%.7E" old_value
        grad_string = @sprintf "%.7E" gradient[j]
        alpha_p = (prob.state_vector[j].value - old_value)/gradient[j]
        alpha_string = @sprintf "%.7E" alpha_p
        step = gradient[j] * alpha_p
        sigma_string = @sprintf "%.7E" step

        write(fid,istring*"; "*jstring*"; ")
        write(fid,ov_string*"; "*nv_string*"; "*grad_string*"; "*alpha_string*"; "*sigma_string*"\n")
    end
    close(fid)
end